import unittest
from pathlib import Path

from ..config import Config
from ..sorting import SortableImport, usort_string
from ..util import try_parse

DEFAULT_CONFIG = Config()


class BasicOrderingTest(unittest.TestCase):
    maxDiff = None

    def test_order(self) -> None:
        items_in_order = [
            b"from __future__ import division",
            b"import os",
            b"from os import path",
            b"import tp",
            b"from tp import x",
            b"from .. import c",
            b"from . import a",
            b"from . import b",
            b"from .a import z",
        ]

        nodes = [
            SortableImport.from_node(
                try_parse(Path("test.py"), data=x).body[0], config=DEFAULT_CONFIG
            )
            for x in items_in_order
        ]
        self.assertEqual(nodes, sorted(nodes))


class UsortStringFunctionalTest(unittest.TestCase):
    def test_sort_ordering(self) -> None:
        # This only tests ordering, not any of the comment or whitespace
        # modifications.
        self.assertEqual(
            """\
import a
import a.b
import b
from a import foo
from a.b import foo2
""",
            usort_string(
                """\
import a
import a.b
from a.b import foo2
from a import foo
import b
""",
                DEFAULT_CONFIG,
            ),
        )

    def test_sort_blocks(self) -> None:
        # This only tests that there are two blocks and we only reorder within a
        # block
        self.assertEqual(
            """\
import c
import d
print("hi")
import a
import b
""",
            usort_string(
                """\
import d
import c
print("hi")
import b
import a
""",
                DEFAULT_CONFIG,
            ),
        )

    # Disabled until wrapping is supported
    #     def test_sort_wrap_moves_comments(self):
    #         # Test that end-of-line directive comments get moved to the first line
    #         # when wrapping is going to happen.
    #         self.assertEqual(
    #             """\
    # from a (  # pylint: disable=E1
    #     import foo,
    # )
    # """,
    #             usort_string(
    #                 """\
    # from a import foo  # pylint: disable=E1
    # """,
    #                 line_limit=10,
    #             ),
    #         )

    def test_shadowed_import(self) -> None:
        # Test that a new block is started when there's a duplicate name
        self.assertEqual(
            """\
import b as b
import a as b
""",
            usort_string(
                """\
import b as b
import a as b
""",
                DEFAULT_CONFIG,
            ),
        )

    def test_dot_handling(self) -> None:
        # Test that 'from .. import b' comes before 'from ..a import foo'
        self.assertEqual(
            """\
from .. import b
from ..a import foo
from . import d
from .c import e
""",
            usort_string(
                """\
from ..a import foo
from .. import b
from . import d
from .c import e
""",
                DEFAULT_CONFIG,
            ),
        )


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from dataclasses import dataclass, field
from typing import List, Optional, Set

import libcst as cst

from .config import Config
from .util import try_parse


def with_dots(x: cst.CSTNode) -> str:
    if isinstance(x, cst.Attribute):
        return ".".join([with_dots(x.value), with_dots(x.attr)])
    elif isinstance(x, cst.Name):
        return x.value
    else:
        raise TypeError(f"Can't with_dots on {type(x)}")


@dataclass(order=True)
class SortKey:
    category_index: int
    is_from_import: bool
    ndots: int


@dataclass(order=True)
class SortableImport:
    node: cst.CSTNode = field(repr=False, compare=False)
    sort_key: SortKey = field(init=False)

    # For constructing the sort key...
    first_module: str
    first_dotted_import: str

    config: Config = field(repr=False, compare=False)

    # This is only used for detecting unsafe ordering, and is not used for
    # breaking ties.  e.g. `import a as b; import b.c` shadows `b`.
    imported_names: Set[str] = field(default_factory=set, compare=False)

    def __post_init__(self) -> None:
        ndots = len(self.first_module) - len(self.first_module.lstrip("."))
        assert isinstance(self.node, cst.SimpleStatementLine)
        self.sort_key = SortKey(
            # TODO this will raise on missing category
            category_index=self.config.categories.index(
                self.config.category(self.first_module)
            ),
            is_from_import=isinstance(self.node.body[0], cst.ImportFrom),
            # replicate ... sorting before .. before .
            ndots=-ndots,
        )

    @classmethod
    def from_node(cls, node: cst.CSTNode, config: Config) -> "SortableImport":
        # TODO: This duplicates (differently) what's in the LibCST import
        # metadata visitor.  This is also a bit hard to read.
        if isinstance(node, cst.SimpleStatementLine):
            first_module: Optional[str] = None
            first_dotted_import: Optional[str] = None
            names: List[str] = []
            sort_key: Optional[str] = None

            # There are 4 basic types of import
            # Additionally some forms z can have leading dots for relative
            # imports, and there can be multiple on the right-hand side.
            #
            if isinstance(node.body[0], cst.Import):
                # import z
                # import z as y
                for name in node.body[0].names:
                    if name.asname:
                        names.append(with_dots(name.asname.name))
                    else:
                        names.append(with_dots(name.name).split(".")[0])

                    if first_module is None:
                        first_module = with_dots(name.name)
                        first_dotted_import = first_module

            elif isinstance(node.body[0], cst.ImportFrom):
                # from z import x
                # from z import x as y

                # This is treated as a barrier and should never get this far.
                assert not isinstance(node.body[0].names, cst.ImportStar)

                if node.body[0].module is None:
                    # from . import foo [as bar]
                    # (won't have dots but with_dots makes the typing easier)
                    sort_key = with_dots(node.body[0].names[0].name)
                else:
                    # from x import foo [as bar]
                    sort_key = with_dots(node.body[0].module)

                if node.body[0].relative:
                    first_dotted_import = sort_key
                    sort_key = "." * len(node.body[0].relative)
                    if node.body[0].module is not None:
                        sort_key += first_dotted_import

                if first_module is None:
                    first_module = sort_key
                    if first_dotted_import is None:
                        for alias in node.body[0].names:
                            first_dotted_import = with_dots(alias.name)
                            break

                for alias in node.body[0].names:
                    if alias.asname:
                        assert isinstance(alias.asname.name, cst.Name)
                        names.append(alias.asname.name.value)
                    else:
                        assert isinstance(alias.name, cst.Name)
                        names.append(alias.name.value)
            else:
                raise TypeError

            assert first_module is not None
            assert first_dotted_import is not None
            return cls(
                node=node,
                first_module=first_module,
                first_dotted_import=first_dotted_import,
                imported_names=set(names),
                config=config,
            )
        raise ValueError("Not an import")


@dataclass
class SortableBlock:
    start_idx: int
    end_idx: Optional[int] = None  # half-open interval

    stmts: List[SortableImport] = field(default_factory=list)
    imported_names: Set[str] = field(default_factory=set)


# TODO can this generalize to functions as well?
def sortable_blocks(mod: cst.Module, config: Config) -> List[SortableBlock]:
    # Finds blocks of imports separated by barriers (non-import statements, or
    # dangerous imports).  We will only sort within a block, and only when there
    # are no duplicate names.

    ret: List[SortableBlock] = []
    cur: Optional[SortableBlock] = None
    for i, stmt in enumerate(mod.body):
        # print(stmt)
        # TODO support known_side_effect_modules or so
        if is_sortable_import(stmt):
            imp = SortableImport.from_node(stmt, config)
            if cur is None:
                cur = SortableBlock(i, i + 1)
                ret.append(cur)

            if cur.imported_names & imp.imported_names:
                # This overwrites an earlier name
                # TODO `os` vs `os.path` shouldn't cause a separate block, but
                # will require a more complex data structure.
                cur = SortableBlock(i, i + 1)
                ret.append(cur)

            cur.end_idx = i + 1
            cur.stmts.append(imp)
            cur.imported_names |= imp.imported_names
        else:
            if cur:
                cur = None
    return ret


def is_sortable_import(stmt: cst.CSTNode) -> bool:
    if isinstance(stmt, cst.SimpleStatementLine):
        # N.b. `body` is a list, because the SimpleStatementLine might have
        # semicolons.  We only look at the first, which is probably the most
        # dangerous thing in here.
        #
        # If black is run, it will put the semicolon pieces on different lines,
        # but we don't want to do that until we reflow and handle directives
        # like noqa.  TODO do that before calling is_sortable_import, and assert
        # that it's not a compound statement line.
        if isinstance(stmt.body[0], cst.ImportFrom):
            # `from x import *` is a barrier
            if isinstance(stmt.body[0].names, cst.ImportStar):
                return False
            return True
        elif isinstance(stmt.body[0], cst.Import):
            return True
        else:
            return False
    else:
        return False


def usort_string(data: str, config: Config) -> str:
    # Intended for unittesting
    mod = try_parse(data=data.encode(), path="<test>")
    blocks = sortable_blocks(mod, config=config)
    for b in blocks:
        sorted_stmts = sorted(b.stmts)
        mod.body[b.start_idx : b.end_idx] = [s.node for s in sorted_stmts]
    return mod.code  # type: ignore



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import enum
from dataclasses import dataclass, field
from pathlib import Path
from typing import Sequence, Set

import toml

from .stdlibs import STDLIB_TOP_LEVEL_NAMES


# TODO these numbers are meaningless, use strings or auto?
class Category(enum.Enum):
    FUTURE = "future"
    STANDARD_LIBRARY = "standard_library"
    FIRST_PARTY = "first_party"
    THIRD_PARTY = "third_party"


@dataclass
class Config:
    known_future: Set[str] = field(default_factory=lambda: {"__future__"})
    known_first_party: Set[str] = field(default_factory=set)
    known_third_party: Set[str] = field(default_factory=set)
    known_standard_library: Set[str] = field(
        default_factory=STDLIB_TOP_LEVEL_NAMES.copy
    )

    # These names are vaguely for compatibility with isort; however while it has
    # separate sections for "current project" and "explicitly local", these are
    # handled differently as "first_party".
    categories: Sequence[Category] = (
        Category.FUTURE,
        Category.STANDARD_LIBRARY,
        Category.THIRD_PARTY,
        Category.FIRST_PARTY,
    )
    default_section: Category = Category.THIRD_PARTY

    @classmethod
    def find(cls, filename: Path) -> "Config":
        rv = cls()

        # TODO This logic should be split out to a separate project, as it's
        # reusable and deserves a number of tests to get right.  Can probably
        # also stop once finding a .hg, .git, etc
        p = filename
        while True:
            if p.is_dir():
                candidate = p / "pyproject.toml"
                if candidate.exists():
                    rv.update_from_config(candidate)
                    break

            # Stop on root (hopefully works on Windows)
            if p.parent == p:
                break
            # Stop on different volume
            if p.exists() and p.stat().st_dev != p.parent.stat().st_dev:
                break

            p = p.parent

        return rv

    def update_from_config(self, toml_path: Path) -> None:
        conf = toml.loads(toml_path.read_text())
        tbl = conf.get("tool", {}).get("usort", {})
        if "known_first_party" in tbl:
            self.known_first_party.update(tbl["known_first_party"])
        if "known_third_party" in tbl:
            self.known_third_party.update(tbl["known_third_party"])
        if "known_standard_library" in tbl:
            self.known_standard_library.update(tbl["known_standard_library"])
        if "categories" in tbl:
            self.categories = [Category(x) for x in tbl["categories"]]
        if "default_section" in tbl:
            self.default_section = Category(tbl["default_section"])

    def update_from_flags(
        self,
        known_first_party: str,
        known_third_party: str,
        known_standard_library: str,
        categories: str,
        default_section: str,
    ) -> None:
        if known_first_party:
            self.known_first_party.update(known_first_party.split(","))
        if known_third_party:
            self.known_third_party.update(known_third_party.split(","))
        if known_standard_library:
            self.known_standard_library.update(known_standard_library.split(","))
        if categories:
            self.categories = [Category(x) for x in categories.split(",")]
        if default_section:
            self.default_section = Category(default_section)

    def category(self, dotted_import: str) -> Category:
        """
        Given a piece of an import string, return its category for this config.

        You can pass in ".foo" or "pkg.foo.bar" or just "os" and it should
        categorize.
        """
        first_part = dotted_import.split(".")[0]
        if first_part == "":
            # relative import
            return Category.FIRST_PARTY
        elif first_part in self.known_future:
            return Category.FUTURE
        elif first_part in self.known_first_party:
            return Category.FIRST_PARTY
        elif first_part in self.known_standard_library:
            return Category.STANDARD_LIBRARY
        elif first_part in self.known_third_party:
            return Category.THIRD_PARTY
        else:
            return self.default_section



import unittest

import libcst as cst

from ..sorting import Config, SortableImport


class SortableImportTest(unittest.TestCase):
    def test_from_node_Import(self) -> None:
        imp = SortableImport.from_node(cst.parse_statement("import a"), Config())
        self.assertEqual("a", imp.first_module)
        self.assertEqual("a", imp.first_dotted_import)
        self.assertEqual({"a"}, imp.imported_names)

        imp = SortableImport.from_node(cst.parse_statement("import a, b"), Config())
        self.assertEqual("a", imp.first_module)
        self.assertEqual("a", imp.first_dotted_import)
        self.assertEqual({"a", "b"}, imp.imported_names)

        imp = SortableImport.from_node(cst.parse_statement("import a as b"), Config())
        self.assertEqual("a", imp.first_module)
        self.assertEqual("a", imp.first_dotted_import)
        self.assertEqual({"b"}, imp.imported_names)

        imp = SortableImport.from_node(cst.parse_statement("import os.path"), Config())
        self.assertEqual("os.path", imp.first_module)
        self.assertEqual("os.path", imp.first_dotted_import)
        self.assertEqual({"os"}, imp.imported_names)

    def test_from_node_ImportFrom(self) -> None:
        imp = SortableImport.from_node(cst.parse_statement("from a import b"), Config())
        self.assertEqual("a", imp.first_module)
        self.assertEqual("b", imp.first_dotted_import)
        self.assertEqual({"b"}, imp.imported_names)

        imp = SortableImport.from_node(
            cst.parse_statement("from a import b as c"), Config()
        )
        self.assertEqual("a", imp.first_module)
        self.assertEqual("b", imp.first_dotted_import)
        self.assertEqual({"c"}, imp.imported_names)

    def test_from_node_ImportFrom_relative(self) -> None:
        imp = SortableImport.from_node(
            cst.parse_statement("from .a import b"), Config()
        )
        self.assertEqual(".a", imp.first_module)
        self.assertEqual("a", imp.first_dotted_import)
        self.assertEqual({"b"}, imp.imported_names)

        imp = SortableImport.from_node(
            cst.parse_statement("from ...a import b"), Config()
        )
        self.assertEqual("...a", imp.first_module)
        self.assertEqual("a", imp.first_dotted_import)
        self.assertEqual({"b"}, imp.imported_names)

        imp = SortableImport.from_node(cst.parse_statement("from . import a"), Config())
        self.assertEqual(".", imp.first_module)
        self.assertEqual("a", imp.first_dotted_import)
        self.assertEqual({"a"}, imp.imported_names)

        imp = SortableImport.from_node(
            cst.parse_statement("from .. import a"), Config()
        )
        self.assertEqual("..", imp.first_module)
        self.assertEqual("a", imp.first_dotted_import)
        self.assertEqual({"a"}, imp.imported_names)

        imp = SortableImport.from_node(
            cst.parse_statement("from . import a as b"), Config()
        )
        self.assertEqual(".", imp.first_module)
        self.assertEqual("a", imp.first_dotted_import)
        self.assertEqual({"b"}, imp.imported_names)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import sys
from pathlib import Path
from typing import List

import click
from moreorless.click import echo_color_unified_diff

from .config import Config
from .sorting import sortable_blocks, try_parse, usort_string
from .util import walk


@click.group()
def main() -> None:
    pass


@main.command()
@click.option("--multiples", is_flag=True, help="Only show files with multiple blocks")
@click.option("--debug", is_flag=True, help="Show internal information")
@click.argument("filenames", nargs=-1)
def list_imports(multiples: bool, debug: bool, filenames: List[str]) -> None:
    # This is used to debug the sort keys on the various lines, and understand
    # where the barriers are that produce different blocks.

    for f in filenames:
        config = Config.find(Path(f))
        mod = try_parse(Path(f))
        try:
            blocks = sortable_blocks(mod, config)
        except Exception as e:
            print("Exception", f, e)
            continue

        if multiples and len(blocks) < 2:
            continue

        click.secho(f"{f} {len(blocks)} blocks:", fg="yellow")
        for b in blocks:
            print(f"  body[{b.start_idx}:{b.end_idx}]")
            sorted_stmts = sorted(b.stmts)
            if debug:
                for s in b.stmts:
                    print(f"    {sorted_stmts.index(s)} {s}")
            else:
                print("Formatted:")
                print("[[[")
                for s in sorted_stmts:
                    print(mod.code_for_node(s.node), end="")
                print("]]]")


@main.command()
@click.option("--diff", is_flag=True)
@click.option("--check", is_flag=True)
@click.argument("filenames", nargs=-1)
def format(diff: bool, check: bool, filenames: List[str]) -> None:
    """
    This is intended to sort nodes separated by barriers, and that's it.
    We don't format them (aside from moving comments).  Black does the rest.
    When in doubt leave lines alone.
    """
    if not filenames:
        raise click.ClickException("Provide some filenames")

    rc = 0
    for f in filenames:
        pf = Path(f)
        config = Config.find(pf)
        if pf.is_dir():
            files = walk(pf, "*.py")
        else:
            files = [pf]

        for pf in files:
            try:
                data = pf.read_text()
                result = usort_string(data, config)
            except Exception as e:
                print(repr(e))
                rc |= 1
                continue
            if diff:
                echo_color_unified_diff(data, result, pf.as_posix())
            elif check:
                if data != result:
                    rc |= 2
                    print(f"Would sort {pf}")
            elif result != data:
                print(f"Sorted {pf}")
                pf.write_text(result)

    sys.exit(rc)


if __name__ == "__main__":
    main()



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Set

from stdlib_list.base import short_versions, stdlib_list

STDLIB_MODULES: Set[str] = set(sum((stdlib_list(v) for v in short_versions), []))

STDLIB_TOP_LEVEL_NAMES: Set[str] = set(v.split(".")[0] for v in STDLIB_MODULES)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
from fnmatch import fnmatch
from pathlib import Path
from typing import Generator, Optional

import libcst as cst


def walk(path: Path, glob: str) -> Generator[Path, None, None]:
    for root, dirs, files in os.walk(path):
        dirs[:] = [d for d in dirs if not d.startswith(".")]
        root_path = Path(root)
        for f in files:
            if fnmatch(f, glob):
                yield root_path / f


def try_parse(path: Path, data: Optional[bytes] = None,) -> cst.Module:
    """
    Attempts to parse the file with all syntax versions known by LibCST.

    If none parse, raises an exception that tells you that (what we know, not an
    error that might not be the most helpful).
    """
    if data is None:
        data = path.read_bytes()

    for version in cst.KNOWN_PYTHON_VERSION_STRINGS:
        try:
            mod = cst.parse_module(
                data, cst.PartialParserConfig(python_version=version)
            )
            return mod
        except cst.ParserSyntaxError:
            continue

    # Intentionally not raising an exception with a specific syntax error (if we
    # keep the last, meaning oldest python version, then it might complain about
    # typehints like https://github.com/psf/black/issues/1158)
    raise Exception(f"No version could parse {path}")



import unittest

from .config import ConfigTest  # noqa: F401
from .functional import BasicOrderingTest, UsortStringFunctionalTest  # noqa: F401
from .sort_key import SortableImportTest  # noqa: F401

unittest.main()



import tempfile
import unittest
from pathlib import Path

from usort.config import Category, Config


class ConfigTest(unittest.TestCase):
    def test_third_party(self) -> None:
        with tempfile.TemporaryDirectory() as d:
            # Defaults should not know about this module.
            # This is a silly value for default_category but this makes it more
            # obvious when it's in use.
            (Path(d) / "pyproject.toml").write_text(
                """\
[tool.usort]
default_section = "future"
"""
            )
            conf = Config.find(Path(d))
            self.assertEqual(set(), conf.known_third_party)
            self.assertEqual(Category.FUTURE, conf.category("psutil.cpu_times"))

            (Path(d) / "pyproject.toml").write_text(
                """\
[tool.usort]
default_section = "future"
known_third_party = ["psutil", "cocoa"]
"""
            )

            conf = Config.find(Path(d))
            self.assertEqual({"psutil", "cocoa"}, conf.known_third_party)
            self.assertEqual(Category.THIRD_PARTY, conf.category("psutil.cpu_times"))

            # Works even on invalid children
            conf = Config.find(Path(d) / "foo")
            self.assertEqual({"psutil", "cocoa"}, conf.known_third_party)
            self.assertEqual(Category.THIRD_PARTY, conf.category("psutil.cpu_times"))
